@@ -8,6 +8,7 @@ use pyo3::{PyMappingProtocol, PyObjectProtocol};
 use tk::normalizer::{char_to_bytes, NormalizedString, Range, SplitDelimiterBehavior};
 use tk::pattern::Pattern;
 
+/// Represents a Pattern as used by `NormalizedString`
 #[derive(Clone, FromPyObject)]
 pub enum PyPattern<'p> {
     #[pyo3(annotation = "str")]
@@ -44,6 +45,15 @@ impl From<PyPattern<'_>> for tk::normalizers::replace::ReplacePattern {
     }
 }
 
+impl From<PyPattern<'_>> for tk::pre_tokenizers::split::SplitPattern {
+    fn from(pattern: PyPattern<'_>) -> Self {
+        match pattern {
+            PyPattern::Str(s) => Self::String(s.to_owned()),
+            PyPattern::Regex(r) => Python::with_gil(|py| Self::Regex(r.borrow(py).pattern.clone())),
+        }
+    }
+}
+
 #[derive(Debug, Clone, FromPyObject)]
 pub enum PyRange<'s> {
     #[pyo3(annotation = "int")]
@@ -173,6 +183,15 @@ fn slice(
         .flatten())
 }
 
+/// NormalizedString
+///
+/// A NormalizedString takes care of modifying an "original" string, to obtain a "normalized" one.
+/// While making all the requested modifications, it keeps track of the alignment information
+/// between the two versions of the string.
+///
+/// Args:
+///     sequence: str:
+///         The string sequence used to initialize this NormalizedString
 #[pyclass(module = "tokenizers", name=NormalizedString)]
 #[derive(Clone)]
 pub struct PyNormalizedString {
@@ -186,6 +205,7 @@ impl PyNormalizedString {
         NormalizedString::from(s).into()
     }
 
+    /// The normalized part of the string
     #[getter]
     fn get_normalized(&self) -> &str {
         self.normalized.get()
@@ -196,70 +216,119 @@ impl PyNormalizedString {
         self.normalized.get_original()
     }
 
+    /// Runs the NFD normalization
+    #[text_signature = "(self)"]
     fn nfd(&mut self) {
         self.normalized.nfd();
     }
 
+    /// Runs the NFKD normalization
+    #[text_signature = "(self)"]
     fn nfkd(&mut self) {
         self.normalized.nfkd();
     }
 
+    /// Runs the NFC normalization
+    #[text_signature = "(self)"]
     fn nfc(&mut self) {
         self.normalized.nfc();
     }
 
+    /// Runs the NFKC normalization
+    #[text_signature = "(self)"]
     fn nfkc(&mut self) {
         self.normalized.nfkc();
     }
 
+    /// Lowercase the string
+    #[text_signature = "(self)"]
     fn lowercase(&mut self) {
         self.normalized.lowercase();
     }
 
+    /// Uppercase the string
+    #[text_signature = "(self)"]
     fn uppercase(&mut self) {
         self.normalized.uppercase();
     }
 
+    /// Prepend the given sequence to the string
+    #[text_signature = "(self, s)"]
     fn prepend(&mut self, s: &str) {
         self.normalized.prepend(s);
     }
 
+    /// Append the given sequence to the string
+    #[text_signature = "(self, s)"]
     fn append(&mut self, s: &str) {
         self.normalized.append(s);
     }
 
+    /// Strip the left of the string
+    #[text_signature = "(self)"]
     fn lstrip(&mut self) {
         self.normalized.lstrip();
     }
 
+    /// Strip the right of the string
+    #[text_signature = "(self)"]
     fn rstrip(&mut self) {
         self.normalized.rstrip();
     }
 
+    /// Strip both ends of the string
+    #[text_signature = "(self)"]
     fn strip(&mut self) {
         self.normalized.strip();
     }
 
+    /// Clears the string
+    #[text_signature = "(self)"]
     fn clear(&mut self) {
         self.normalized.clear();
     }
 
+    /// Slice the string using the given range
+    #[text_signature = "(self, range)"]
     fn slice(&self, range: PyRange) -> PyResult<Option<PyNormalizedString>> {
         slice(&self.normalized, &range)
     }
 
+    /// Filter each character of the string using the given func
+    #[text_signature = "(self, func)"]
     fn filter(&mut self, func: &PyAny) -> PyResult<()> {
         filter(&mut self.normalized, func)
     }
 
+    /// Calls the given function for each character of the string
+    #[text_signature = "(self, func)"]
     fn for_each(&self, func: &PyAny) -> PyResult<()> {
         for_each(&self.normalized, func)
     }
 
+    /// Calls the given function for each character of the string
+    ///
+    /// Replaces each character of the string using the returned value. Each
+    /// returned value **must** be a str of length 1 (ie a character).
+    #[text_signature = "(self, func)"]
     fn map(&mut self, func: &PyAny) -> PyResult<()> {
         map(&mut self.normalized, func)
     }
 
+    /// Split the NormalizedString using the given pattern and the specified behavior
+    ///
+    /// Args:
+    ///     pattern: Pattern:
+    ///         A pattern used to split the string. Usually a string or a Regex
+    ///
+    ///     behavior: SplitDelimiterBehavior:
+    ///         The behavior to use when splitting.
+    ///         Choices: "removed", "isolated", "merged_with_previous", "merged_with_next",
+    ///         "contiguous"
+    ///
+    /// Returns:
+    ///     A list of NormalizedString, representing each split
+    #[text_signature = "(self, pattern, behavior)"]
     fn split(
         &mut self,
         pattern: PyPattern,
@@ -272,6 +341,15 @@ impl PyNormalizedString {
             .collect())
     }
 
+    /// Replace the content of the given pattern with the provided content
+    ///
+    /// Args:
+    ///     pattern: Pattern:
+    ///         A pattern used to match the string. Usually a string or a Regex
+    ///
+    ///     content: str:
+    ///         The content to be used as replacement
+    #[text_signature = "(self, pattern, content)"]
     fn replace(&mut self, pattern: PyPattern, content: &str) -> PyResult<()> {
         ToPyResult(self.normalized.replace(pattern, content)).into()
     }
@@ -333,6 +411,20 @@ impl PyNormalizedStringRefMut {
     pub fn destroyed_error() -> PyErr {
         exceptions::PyException::new_err("Cannot use a NormalizedStringRefMut outside `normalize`")
     }
+
+    /// Provides a way to access a reference to the underlying NormalizedString
+    pub fn map_as_ref<F: FnOnce(&NormalizedString) -> U, U>(&self, f: F) -> PyResult<U> {
+        self.inner
+            .map(f)
+            .ok_or_else(PyNormalizedStringRefMut::destroyed_error)
+    }
+
+    /// Provides a way to access a mutable reference to the underlying NormalizedString
+    pub fn map_as_mut<F: FnOnce(&mut NormalizedString) -> U, U>(&mut self, f: F) -> PyResult<U> {
+        self.inner
+            .map_mut(f)
+            .ok_or_else(PyNormalizedStringRefMut::destroyed_error)
+    }
 }
 
 #[pymethods]
