@@ -1,5 +1,8 @@
-use crate::models::unigram::lattice::Lattice;
-use crate::models::unigram::trie::{Trie, TrieBuilder};
+use super::{
+    lattice::Lattice,
+    trainer::UnigramTrainer,
+    trie::{Trie, TrieBuilder},
+};
 use crate::tokenizer::{Model, Result, Token};
 use crate::utils::cache::Cache;
 
@@ -404,8 +407,10 @@ impl<'a> Iterator for UnigramIterator<'a> {
 }
 
 impl Model for Unigram {
-    fn get_vocab(&self) -> &HashMap<String, u32> {
-        &self.token_to_ids
+    type Trainer = UnigramTrainer;
+
+    fn get_vocab(&self) -> HashMap<String, u32> {
+        self.token_to_ids.clone()
     }
 
     fn get_vocab_size(&self) -> usize {
@@ -433,11 +438,8 @@ impl Model for Unigram {
         self.token_to_ids.get(token).copied()
     }
 
-    fn id_to_token(&self, id: u32) -> Option<&str> {
-        match self.vocab.get(id as usize) {
-            Some(item) => Some(&item.0),
-            None => None,
-        }
+    fn id_to_token(&self, id: u32) -> Option<String> {
+        self.vocab.get(id as usize).map(|item| item.0.clone())
     }
 
     fn save(&self, folder: &Path, name: Option<&str>) -> Result<Vec<PathBuf>> {
@@ -452,6 +454,10 @@ impl Model for Unigram {
         std::fs::write(&fullpath, string)?;
         Ok(vec![fullpath])
     }
+
+    fn get_trainer(&self) -> Self::Trainer {
+        UnigramTrainer::default()
+    }
 }
 
 #[cfg(test)]
